import time, traceback, sys
from .core import Primitive, PyThread, synchronized
from .dequeue import DEQueue
from threading import Timer

class TaskQueueDelegate(object):
    
    # abstract class
    
    def taskIsStarting(self, queue, task):
        pass

    def taskStarted(self, queue, task):
        pass

    def taskEnded(self, queue, task, result):
        pass

    def taskFailed(self, queue, task, exc_type, exc_value, tback):
        pass

class Task(Primitive):

    def __init__(self, name=None):
        Primitive.__init__(self, name=name)
        self.Created = time.time()
        self.Queued = None
        self.Started = None
        self.Ended = None
    
    #def __call__(self):
    #    pass

    def run(self):
        raise NotImplementedError
        
    @property
    def has_started(self):
        return self.Started is not None
        
    @synchronized
    @property
    def is_running(self):
        return self.Started is not None and self.Ended is None
        
    @synchronized
    @property
    def has_ended(self):
        return self.Started is not None and self.Ended is not None
        
    def start(self):
        self.Started = time.time()
        
    def end(self):
        self.Ended = time.time()
        
    def enqueue(self):
        self.Queued = time.time()

class FunctionTask(Task):

    def __init__(self, fcn, *params, **args):
        Task.__init__(self)
        self.F = fcn
        self.Params = params
        self.Args = args
        
    def run(self):
        return self.F(*self.Params, **self.Args)
        
class TaskQueue(Primitive):
    
    class ExecutorThread(PyThread):
        def __init__(self, queue, task):
            PyThread.__init__(self, daemon=True)
            self.Queue = queue
            self.Task = task
            
        def run(self):
            task = self.Task
            task.start()
            try:
                if callable(task):
                    result = task()
                else:
                    result = task.run()
                task.end()
                self.Queue.taskEnded(self.Task, result)
            except:
                task.end()
                exc_type, value, tb = sys.exc_info()
                self.Queue.taskFailed(self.Task, exc_type, value, tb)
            finally:
                self.Queue.threadEnded(self)
                self.Queue = None
                    
    def __init__(self, nworkers, capacity=None, stagger=None, tasks = [], delegate=None, name=None):
        Primitive.__init__(self, name=name)
        self.NWorkers = nworkers
        self.Threads = []
        self.Queue = DEQueue(capacity)
        self.Held = False
        self.Stagger = stagger or 0.0
        self.LastStart = 0.0
        self.StartTimer = None
        self.Delegate = delegate
        for t in tasks:
            self.addTask(t)

    def addTask(self, task, timeout = None):
        #print "addTask() entry"
        self.Queue.append(task, timeout=timeout)
        #print("queue.append done", self.counts())
        task.enqueue()
        self.startThreads()
        return self
        
    def __iadd__(self, task):
        return self.addTask(task)
        
    def __lshift__(self, task):
        return self.addTask(task)
        
    def insertTask(self, task, timeout = None):
        self.Queue.insert(task, timeout = timeout)
        task.enqueue()
        self.startThreads()
        return self
        
    def __rshift__(self, task):
        return self.insertTask(task)
        
    @synchronized
    def timer_fired(self):
        self.StartTimer = None
        self.startThreads()
        
    @synchronized
    def startThreads(self):
        #print "startThreads() entry"
        if not self.Held:
            while self.Queue \
                    and (self.NWorkers is None or len(self.Threads) < self.NWorkers) \
                    and not self.Held:
                if self.Stagger > 0.0:
                    delta = self.LastStart + self.Stagger - time.time()
                    #print "arming timer..."
                    if delta > 0.0:
                        if self.StartTimer is None:
                            self.StartTimer = Timer(delta, self.timer_fired)
                            self.StartTimer.daemon = True
                            self.StartTimer.start()
                        break
                self.LastStart = time.time()
                task = self.Queue.pop()
                t = self.ExecutorThread(self, task)
                t.kind = "%s.task" % (self.kind,)
                self.Threads.append(t)
                self.call_delegate("taskIsStarting", self, task)
                t.start()
                self.call_delegate("taskStarted", self, task)
                #print("thread started")
            
    @synchronized
    def threadEnded(self, t):
        #print("queue.threadEnded: ", t)
        if t in self.Threads:
            self.Threads.remove(t)
        self.startThreads()
        self.wakeup()
        
    def call_delegate(self, cb, *params):
        if self.Delegate is not None and hasattr(self.Delegate, cb):
            try:    getattr(self.Delegate, cb)(*params)
            except:
                traceback.print_exc(file=sys.stderr)
            
    def taskEnded(self, task, result):
        self.call_delegate("taskEnded", self, task, result)
        
    def taskFailed(self, task, exc_type, exc_value, tb):
        if self.Delegate is None:
            sys.stdout.write("Exception in task %s:\n" % (task, ))
            traceback.print_exception(exc_type, exc_value, tb, file=sys.stderr)
        else:
            self.call_delegate("taskFailed", self, task,  exc_type, exc_value, tb)
            
    @synchronized
    def waitingTasks(self):
        return list(self.Queue.items())
        
    @synchronized
    def activeTasks(self):
        return [t.Task for t in self.Threads]
        
    @synchronized
    def tasks(self):
        return self.waitingTasks(), self.activeTasks()
        
    def nrunning(self):
        return len(self.Threads)
        
    def nwaiting(self):
        return len(self.Queue)
        
    @synchronized
    def counts(self):
        return self.nwaiting(), self.nrunning()
        
    @synchronized
    def hold(self):
        self.Held = True
        
    @synchronized
    def release(self):
        self.Held = False
        self.startThreads()
        
    @synchronized
    def isEmpty(self):
        return len(self.Queue) == 0 and len(self.Threads) == 0
                
    def waitUntilEmpty(self):
        # wait until all tasks are done and the queue is empty
        if not self.isEmpty():
            while not self.sleep(function=self.isEmpty):
                pass
                
    join = waitUntilEmpty
                
    def drain(self):
        self.hold()
        self.waitUntilEmpty()
                
    @synchronized
    def flush(self):
        self.Queue.flush()
        self.wakeup()
            
    def __len__(self):
        return len(self.Queue)




from .core import Primitive

class Flag(Primitive):
    
    def __init__(self, value=False):
        Primitive.__init__(self)
        self._Value = value
        
    def set(self, value=True):
        self._Value = value
        self.wakeup()
        
    def get(self):
        return self._Value
        
    value = property(get, set)
        
    def sleep_until(self, value=None, predicate=None, timeout = None):
        if value is not None:
            predicate = lambda x, v=value: x == v
        if predicate is None:
            predicate = lambda x: not not x     # default: wait until the value evaluates to True
        return self.sleep_until(lambda flag: predicate(flag.Value), self, timeout=timeout)
                
            


from .core import Primitive, synchronized, PyThread, gated, TimerThread, Timeout
from .dequeue import DEQueue
from .task_queue import TaskQueue, Task
from .Scheduler import Scheduler
from .Subprocess import Subprocess, ShellCommand
from .RWLock import RWLock
from .Version import Version
from .promise import Promise
from .processor import Processor
from .flag import Flag
from .gate import Gate
from .LogFile import LogFile, LogStream


__version__ = Version
version_info = tuple(Version.split("."))


__a_ll__ = [
    'Primitive',
    'PyThread',
    'TimerThread',
    'DEQueue',
    'gated',
    'synchronized',
    'Task',
    'TaskQueue',
    'Subprocess',
    'ShellCommand',
    'Version', '__version__', 'version_info',
    'Timeout',
    'Promise',
    'Scheduler',
    'Gate', 'LogFile', 'LogStream'
]



from .core import Primitive, synchronized
from threading import current_thread
import threading

def _thread_id():
    return current_thread().ident

class RWLock(Primitive):
    
    def __init__(self):
        Primitive.__init__(self)
        self.Exclusive = None
        self.ExclusiveCount = 0
        self.ExclusiveQueue = []
        self.Shared = {}
        
    def __purge(self):
        threads = set([t.ident for t in threading.enumerate()])
        if not self.Exclusive in threads:
            self.Exclusive = None
            self.ExclusiveCount = None
        for t in list(self.Shared.keys()):
            if not t in threads:
                del self.Shared[t]
        self.ExclusiveQueue = [t for t in self.ExclusiveQueue if t in threads]
    
    def __acq_exclusive(self):
        tid = _thread_id()
        self.__purge()
        if tid == self.Exclusive:
            self.ExclusiveCount += 1
            return True
        if self.Exclusive == None and (len(self.Shared) == 0 or len(self.Shared) == 1 and tid in self.Shared):
            if self.ExclusiveQueue:
                if tid == self.ExclusiveQueue[0]:
                    self.ExclusiveQueue = self.ExclusiveQueue[1:]
                else:
                    self.ExclusiveQueue.append(tid)
                    return False
            self.Exclusive = tid
            self.ExclusiveCount = 1
            return True
        return False
        
    def __rel_exclusive(self):
        tid = _thread_id()
        assert self.Exclusive == tid and self.ExclusiveCount > 0
        self.ExclusiveCount -= 1
        if self.ExclusiveCount <= 0:
            self.ExclusiveCount = 0
            self.Exclusive= None
        self.wakeup(all=True)
        
    def __acq_shared(self):
        tid = _thread_id()
        self.__purge()
        if not self.Exclusive in (None, tid):
            return False
        if not tid in self.Shared:
            self.Shared[tid] = 0
        self.Shared[tid] = self.Shared[tid] + 1
        return True
            
    def __rel_shared(self):
        tid = _thread_id()
        n = self.Shared.get(tid)
        assert n is not None and n > 0
        n -= 1
        if n == 0:
            del self.Shared[tid]
        else:
            self.Shared[tid] = n
        self.wakeup(all=True)
        
    @synchronized
    def acquireExclusive(self):
        while not self.__acq_exclusive():
            self.sleep()
            
    @synchronized
    def acquireShared(self):
        while not self.__acq_shared():
            self.sleep()
            
    @synchronized
    def releaseExclusive(self):
        self.__rel_exclusive()
        
    @synchronized
    def releaseShared(self):
        self.__rel_shared()
        
            
    @property
    def exclusive(self):
        class ExclusiveContext(object):
            def __init__(self, rwlock):
                self.RWLock = rwlock
            
            def __enter__(self):
                return self.RWLock.acquireExclusive()
            
            def __exit__(self, t, v, tb):
                return self.RWLock.releaseExclusive()
        return ExclusiveContext(self)
    
    @property
    def shared(self):
        class SharedContext(object):
            def __init__(self, rwlock):
                self.RWLock = rwlock
            
            def __enter__(self):
                return self.RWLock.acquireShared()
            
            def __exit__(self, t, v, tb):
                return self.RWLock.releaseShared()
            
        return SharedContext(self)
        
    @synchronized
    def owners(self):
        return self.Exclusive, list(self.Shared.keys())
        



import time, random
from pythreader import TaskQueue, Task
from threading import Timer

class MyTask(Task):

    def __init__(self, tid):
        Task.__init__(self)
        self.Id = tid

    def run(self):
        print (time.time(), self.Id, "started as instance")
        time.sleep(random.random()*3)
        print (self.Id, "ended")

    def failed(self, e):
        print (e)

q = TaskQueue(5, stagger=1.0, tasks=[MyTask(x) for x in range(10)])
q << MyTask(30) << MyTask(31)
MyTask(32) >> q
q += MyTask(33)


Timer(3.5, lambda q: q.addTask(MyTask(40)), (q,)).start()
Timer(3.51, lambda q: q.addTask(MyTask(41)), (q,)).start()
Timer(3.52, lambda q: q.addTask(MyTask(42)), (q,)).start()
Timer(3.53, lambda q: q.addTask(MyTask(43)), (q,)).start()
Timer(3.9, lambda q: q.addTask(MyTask(44)), (q,)).start()

q.waitUntilEmpty()





from threading import RLock, Thread, Event, Condition, Semaphore, currentThread, get_ident
import time
import sys

Waiting = []
In = []

class Timeout(Exception):
    pass

def threadName():
    t = currentThread()
    return str(t)

def synchronized(method):
    def smethod(self, *params, **args):
        me = get_ident()
        #print("entering synchronized", self, me)
        with self:
            out = method(self, *params, **args)
        #print("exiting synchronized", self, me)
        return out
    return smethod

def gated(method):
    def smethod(self, *params, **args):
        with self._Gate:
            out = method(self, *params, **args)
        return out
    return smethod


def printWaiting():
    print("waiting:----")
    for w in Waiting:
        print(w)
    print("in:---------")
    for w in In:
        print(w)

class UnlockContext(object):
    
    def __init__(self, prim):
        self.Prim = prim
        
    def __enter__(self):
        self.Prim._Lock.__exit__(None, None, None)
        
    def __exit__(self, exc_type, exc_value, traceback):
        self.Prim._Lock.__enter__()    

class Primitive:
    def __init__(self, gate=1, lock=None, name=None):
        self._Kind = self.__class__.__name__
        self._Lock = lock if lock is not None else RLock()
        #print ("Primitive:", self, " _Lock:", self._Lock)
        self._WakeUp = Condition(self._Lock)
        self._Gate = Semaphore(gate)
        self.Name = name
        
    def __str__(self):
        ident = ('"%s"' % (self.Name,)) if self.Name else ("@%s" % (("%x" % (id(self),))[-4:],))
        return "[%s %s]" % (self._Kind, ident)

    def __get_kind(self):
        return self._Kind

    def __set_kind(self, kind):
        self._Kind = kind

    kind = property(__get_kind, __set_kind)

    def getLock(self):
        return self._Lock
        
    def __enter__(self):
        #t = currentThread()
        #print ">>>entry by thread %s %x: %s %x..." % (t.__class__.__name__, id(t), self.kind, id(self))
        return self._Lock.__enter__()
        
    def __exit__(self, exc_type, exc_value, traceback):
        #t = currentThread()
        #print "<<<<exit by thread %s %x: %s %x" % (t.__class__.__name__, id(t), self.kind, id(self))
        return self._Lock.__exit__(exc_type, exc_value, traceback)
    
    @property
    def unlock(self):
        return UnlockContext(self)

    @synchronized
    def sleep(self, timeout = None, function=None, arguments=()):
        #print("sleep", self, get_ident(), "   condition lock:", self._WakeUp._lock, "...")
        self._WakeUp.wait(timeout)
        if function is not None:
            return function(*arguments)

    @synchronized
    def sleep_until(self, predicate, *params, timeout = None, **args):
        #print("sleep", self, get_ident(), "   condition lock:", self._WakeUp._lock, "...")
        t1 = None if timeout is None else time.time() + timeout
        while (t1 is None or time.time() < t1):
            if predicate(*params, **args):
                return
            delta = None
            if t1 is not None:
                delta = max(0.0, t1 - time.time())
            self.sleep(delta)
        else:
            raise Timeout()
            
    # await is a reserved word in Python 3, use "wakeup" instead
    @synchronized
    def wakeup(self, n=1, all=True, function=None, arguments=()):
        if function is not None:
            function(*arguments)
        if all:
            self._WakeUp.notifyAll()
        else:
            self._WakeUp.notify(n)

if sys.version_info < (3,0):
    # await is a reserved word in Python 3, keep it for backward compatibility
    # in Python 2.
    setattr(Primitive, "await", Primitive.sleep)


class PyThread(Thread, Primitive):
    def __init__(self, *params, name=None, **args):
        Thread.__init__(self, *params, **args)
        Primitive.__init__(self, name=name)

class TimerThread(PyThread):
    def __init__(self, function, interval, *params, **args):
        PyThread.__init__(self, daemon=True)
        self.Interval = interval
        self.Func = function
        self.Params = params
        self.Args = args
        self.Pause = False

    def run(self):
        while True:
            while self.Pause:
                self.sleep()
            self.Func(*self.Params, **self.Args)
            time.sleep(self.Interval)
    
    def pause(self):
        self.Pause = True
        
    def resume(self):
        self.Pause = False
        self.wakeup()
                
            
            



from .core import Primitive, synchronized, Timeout
from threading import get_ident, RLock

class DebugLock(object):
    
    def __init__(self):
        self.R = RLock()
        
    def acquire(self, *params, **args):
        print(self, "acquire by ", get_ident(), "...")
        self.R.acquire(*params, **args)
        print(self, "acquired by ", get_ident())
        
    def release(self):
        print(self, "released by ", get_ident())
        self.R.release()
        
    def __enter__(self):
        return self.acquire()
        
    def __exit__(self, *params):
        return self.release()
    

class Promise(Primitive):
    
    def __init__(self, data=None, callbacks = [], name=None):
        Primitive.__init__(self, name=name)    #, lock=DebugLock())
        self.Data = data
        self.Callbacks = callbacks[:]
        self.Complete = False
        self.Canceled = False
        self.Result = None
        self.ExceptionInfo = None     # or tuple (exc_type, exc_value, exc_traceback)
        self.OnDone = self.OnCancel = self.OnException = None
        self.Chained = []
        self.Name = name
        
    @synchronized
    def ondone(self, cb):
        self.OnDone = cb
        
    @synchronized
    def oncancel(self, cb):
        self.OnCancel = cb
        
    @synchronized
    def oncexception(self, cb):
        self.OnException = cb
        
    @synchronized
    def addCallback(self, cb):
        if self.Complete and not self.Canceled:
            cb(self)
        self.Callbacks.append(cb)        
        
    @synchronized
    def exception(self, exc_type, exc_value, exc_traceback):
        self.ExceptionInfo = (exc_type, exc_value, exc_traceback)
        if self.OnException is not None:
            self.OnExceptipn(self)
        for p in self.Chained:
            p.exception(exc_type, exc_value, exc_traceback)
        self.wakeup()
        self._cleanup()

    @synchronized
    def chain(self, *promises):
        if self.ExceptionInfo:
            exc_type, exc_value, exc_traceback = self.ExceptionInfo
            for p in promises:
                p.exception(exc_type, exc_value, exc_traceback)
        elif self.Complete:
            for p in promises:
                p.complete(self.Result)
        elif self.Canceled:
            for p in promises:
                p.complete(None)
        else:
            self.Chained += list(promises)
        return self
        
    @synchronized
    def complete(self, result=None):
        self.Result = result
        self.Complete = True
        if not self.Canceled:
            if self.OnDone is not None:
                self.OnDone(self.Result,)
            for cb in self.Callbacks:
                if cb(self) == "stop":
                    break
        for p in self.Chained:
            #print("%s: complete(%s) ..." % (self, p))
            p.complete(result)
            #print("%s: complete(%s) done" % (self, p))
        self.wakeup()
        self._cleanup()
    
    def is_complete(self):
        return self.Complete
        
    @synchronized
    def wait(self, timeout=None):
        #print("thread %s: wait(%s)..." % (get_ident(), self))
        pred = lambda x: x.Complete or x.Canceled or self.ExceptionInfo is not None
        self.sleep_until(pred, self, timeout=timeout)
        try:
            if self.Complete:
                return self.Result
            elif self.Canceled:
                return None
            elif self.ExceptionInfo:
                _, e, _ = self.ExceptionInfo
                raise e 
            else:
                raise Timeout()
        finally:
            self._cleanup()
            
            
    def _cleanup(self):
        self.Chained = []
        self.Callbacks = []
        self.OnCancel = self.OnException = self.OnDone = None
    
class ORPromise(Primitive):
    
    def __init__(self, promises):
        Primitive.__init__(self)
        self.Fulfilled = None
        for p in promises:
            p.addCallback(self, promise_callback)

    @synchronized
    def promise_callback(self, promise):
        self.Fulfilled = promise
        self.wakeup()
    
    @synchronized
    def wait(self, timeout = None):
        while self.Fulfilled is None:
            self.sleep(timeout)
        return sef.Fulfilled

class ANDPromise(Primitive):
    
    def __init__(self, promises):
        Primitive.__init__(self)
        self.Promises = promises
        
    def wait(self, timeout=None):
        return [p.wait(timeout) for p in self.Promises]



from .core import Primitive, synchronized
import time

class DEQueue(Primitive):

    def __init__(self, capacity=None):
        Primitive.__init__(self)
        self.Capacity = capacity
        self.List = []
        self.Closed = False
    
    @synchronized
    def close(self):
        self.Closed = True
        self.wakeup()
        
    @synchronized    
    def append(self, item, timeout=None):
        t0 = time.time()
        t1 = None if timeout is None else t0 + timeout
        while self.Capacity is not None and len(self.List) >= self.Capacity \
                        and not self.Closed:
            dt = None
            if t1 is not None:
                t = time.time()
                if t > t1:
                    raise RuntimeError("Operation timed-out")
                dt = t1 - t
            self.sleep(dt)
        if self.Closed:
            raise RuntimeError("Queue is closed")
        self.List.append(item)
        self.wakeup()
        
    def __lshift__(self, item):
        return self.add(item)
    
    @synchronized    
    def insert(self, item, timeout=None):
        assert not self.Closed, "The queue is closed"
        t0 = time.time()
        t1 = None if timeout is None else t0 + timeout
        while self.Capacity is not None and len(self.List) >= self.Capacity:
            dt = None
            if t1 is not None:
                t = time.time()
                if t > t1:
                    raise RuntimeError("Operation timed-out")
                dt = t1 - t
            self.sleep(dt)
        self.List.insert(0, item)
        self.wakeup()

    def __rrshift__(self, item):
        return self.push(item)
        
    @synchronized
    def pop(self):
        while len(self.List) == 0 and not self.Closed:
            self.sleep()
        if len(self.List) == 0:
            return None     # closed
        item = self.List[0]
        self.List = self.List[1:]
        self.wakeup()
        return item

    #
    # Iterator protocol
    # 
    # for item in queue:        # wait for next item to arrive
    #   # ... process item
    #
    
    def __iter__(self):
        return self

    @synchronized
    def __next__(self):
        while not self.Closed or len(self) > 0:
            while not self.Closed and len(self) == 0:
                self.sleep()
            item = self.pop()
            if item is not None:
                return item
            else:
                break
        raise StopIteration()

    next = __next__
        
    @synchronized
    def flush(self):
        self.List = []
        self.wakeup()
        
    @synchronized
    def items(self):
        return self.List[:]
        
    @synchronized
    def look(self):
        return self.List[0] if self.List else None
        
    @synchronized
    def popIfFirst(self, x):
        if self.List:
            first = self.List[0]
            if first is x or first == x:
                self.pop()
                return x
        return None
        
    
    def empty(self):
        return not self.List
        
    def __len__(self):
        return len(self.List)
        



import time
import os
import datetime
from .core import PyThread, synchronized, Primitive

def make_timestamp(t=None):
    if t is None:   
        t = datetime.datetime.now()
    elif isinstance(t, (int, float)):
        t = datetime.datetime.fromtimestamp(t)
    return t.strftime("%m/%d/%Y %H:%M:%S") + ".%03d" % (t.microsecond//1000)
                
class LogStream(Primitive):

    def __init__(self, stream, add_timestamp=True, name=None):
        Primitive.__init__(self, name=name)
        self.Stream = stream            # sys.stdout, sys.stderr
        self.AddTimestamps = add_timestamp

    @synchronized
    def log(self, msg, raw=False, add_timestamp=True):
        if add_timestamp and not raw:
            msg = "%s: %s" % (make_timestamp(), msg)
        self.Stream.write(msg + '\n');
        self.Stream.flush()

    @synchronized
    def write(self, msg):
        self.log(msg, raw=True)
        
class LogFile(PyThread):

        def __init__(self, path, interval = '1d', keep = 10, add_timestamp=True, append=True, flush_interval=5.0,
                    name=None):
                # interval = 'midnight' means roll over at midnight
                PyThread.__init__(self, name=name)
                assert isinstance(path, str)
                self._Lock = RLock()
                self.Path = path
                self.File = None
                self.CurLogBegin = 0
                if type(interval) == type(''):
                        mult = 1
                        if interval[-1] == 'd' or interval[-1] == 'D':
                                interval = interval[:-1]
                                mult = 24 * 3600
                                interval = int(interval) * mult
                        elif interval[-1] == 'h' or interval[-1] == 'H':
                                interval = interval[:-1]
                                mult = 3600
                                interval = int(interval) * mult
                        elif interval[-1] == 'm' or interval[-1] == 'M':
                                interval = interval[:-1]
                                mult = 60
                                interval = int(interval) * mult
                self.Interval = interval
                self.Keep = keep
                self.AddTimestamps = add_timestamp
                self.LineBuf = ''
                self.LastLog = None
                self.LastFlush = time.time()
                self.FlushInterval = flush_interval
                if append:
                    self.File = open(self.Path, 'a')
                    self.File.write("%s: [appending to old log]\n" % (make_timestamp(),))
                    self.CurLogBegin = time.time()
                #print("LogFile: created with file:", self.File)
                    
        def run(self):
            while True:
                time.sleep(self.FlushInterval)
                self.flush()
                    
        def newLog(self):
            if self.File != None:
                    self.File.close()
            try:    os.remove('%s.%d' % (self.Path, self.Keep))
            except: pass
            for i in range(self.Keep - 1):
                    inx = self.Keep - i
                    old = '%s.%d' % (self.Path, inx - 1)
                    new = '%s.%d' % (self.Path, inx)
                    try:    os.rename(old, new)
                    except: pass
            try:    os.rename(self.Path, self.Path + '.1')
            except: pass
            self.File = open(self.Path, 'w')
            self.CurLogBegin = time.time()

        @synchronized
        def log(self, msg, raw=False, add_timestamp=True):
            t = time.time()
            if self.Interval == 'midnight':
                if datetime.date.today() != self.LastLog:
                        self.newLog()
            elif isinstance(self.Interval, (int, float)):
                if t > self.CurLogBegin + self.Interval:
                        self.newLog()
            if add_timestamp and not raw:
                msg = "%s: %s" % (make_timestamp(t), msg)
            self._write(msg if raw else msg + "\n")

        @synchronized
        def write(self, msg):
            self.log(msg, raw=True)
            
        @synchronized
        def _write(self, msg):
            if msg:
                #print("LogFile.write: writing to:", self.File)
                self.File.write(msg)
            self.flush()
            self.LastLog = datetime.date.today()
            
        @synchronized
        def flush(self):
            if time.time() > self.LastFlush + self.FlushInterval:
                self.File.flush()
                self.LastFlush = time.time()
        
        def __del__(self):
            #self.flush()
            if self.File is not None:
                self.File.close()
                self.File = None





from .core import PyThread, synchronized
import time, uuid, traceback
import sys

class Scheduler(PyThread):
    def __init__(self):
        PyThread.__init__(self)
        self.Tasks = {}     # id -> (function, t, interval, param, args)

    @synchronized        
    def add(self, fcn, interval=None, param=None, t0=None, id=None, **args):
        #
        # t0 - first time to run the task. Default:
        #   now + interval or now if interval is None
        # interval - interval to repeat the task. Default: do not repeat
        #
        # fcn:
        #   next_t = fcn()
        #   next_t = fcn(param)
        #   next_t = fcn(**args)
        # 
        #   next_t:
        #       "stop" - remove task
        #       int or float - next time to run
        #       None - run at now+interval next time
        #
        if id is None:
            id = uuid.uuid1().hex
        if t0 is None:
            t0 = time.time() + (interval or 0.0)
        self.Tasks[id] = (fcn, t0, interval, param, args)
        self.wakeup()
        return id
        
    @synchronized
    def remove(self, tid):
        if tid in self.Tasks:
            del self.Tasks[tid]
        
    @synchronized        
    def tick(self):
        #print("tick: tasks:", self.Tasks)
        
        now = time.time()
        tasks_to_run = []
        for tid, (fcn, t, interval, param, args) in list(self.Tasks.items()):
            if t <= now:
                with self.unlock:
                    next_t = None       # default
                    try:
                            if param is not None:
                                next_t = fcn(param)
                            elif args is not None:
                                next_t = fcn(**args)
                            else:
                                next_t = fcn()
                    except:
                        sys.stderr.write(traceback.format_exc())
                    #print("exiting unlock context...")
                #print("tick: next_t=", next_t)
                if tid in self.Tasks:
                    # the task may have been unregistered from within fcn() call
                    if next_t == "stop" or \
                            next_t is None and (interval is None or interval < 0.0):
                        del self.Tasks[tid]
                        continue
                    if next_t is None:
                        next_t = now + interval
                    #print("tick: next_t->", next_t)
                    self.Tasks[tid] = (fcn, next_t, interval, param, args)
                        
    def run(self):
        while True:
            delta = 10.0
            with self:
                if self.Tasks:
                    tmin = min(t for _, t, _, _, _ in self.Tasks.values())
                    now = time.time()
                    delta = tmin - time.time()
            if delta > 0.0:
                self.sleep(delta)
            self.tick()

if __name__ == "__main__":
    from datetime import datetime
    
    s = Scheduler()
    
    class NTimes(object):
        
        def __init__(self, n):
            self.N = n
            
        def __call__(self, message):
            t = datetime.now()
            print("%s: %s" % (strftime(t, "%H:%M:%S.f"), message))
            self.N -= 1
            if self.N <= 0:
                print("stopping", message)
                return "stop"
                
    s.start()
    time.sleep(1.0)
    s.add(NTimes(10), 1.5, "green 1.5 sec")
    s.add(NTimes(7), 0.5, "blue 0.5 sec")

    s.join()

        



from .core import PyThread, synchronized
from .promise import Promise
from threading import Timer
import subprocess

def to_str(b):
    if isinstance(b, str):
        return b
    elif isinstance(b, bytes):
        return b.decode("utf-8")
    else:
        return b

class Subprocess(PyThread):
    
    def __init__(self, *params, tag=None, input=None, daemon=None, timeout=None, **kv):
        PyThread.__init__(self, daemon=daemon)    # timeout will be passed to the run()
        self.Params = params
        self.KV = kv
        self.Killed = False
        self.Subrocess = None
        self.Daemon = daemon
        self.Promise = None
        self.Input = input
        self.Timeout = timeout
        self.Tag = tag          # used to identify the Promise
        
    def run(self):
        timeout = self.Timeout
        #print("Subprocess.run(): timeout:", timeout)
        kv = self.KV.copy()
        kv["stderr"] = subprocess.PIPE
        kv["stdout"] = subprocess.PIPE
        
        self.Subprocess = subprocess.Popen(*self.Params, **kv)
        if self.Input is not None:
            self.Subprocess.stdin.write(self.Input)

        killer = None
        if timeout is not None:
            killer = Timer(timeout, self.killme)
            killer.start()          # do not start killer until self.Subprocess is set
            #print("killer started")

        out, err = self.Subprocess.communicate()
        out = to_str(out)
        err = to_str(err)
        retcode = self.Subprocess.returncode

        with self:
                # make this a critical section so the killer process does not intercept us
                if killer is not None:  
                    try:    killer.cancel()
                    except: pass
                self.Subprocess = None
                
        if self.Killed:
            exc = RuntimeError("timeout")
            if self.Promise is not None:
                self.Promise.exception(RuntimeError, exc, None)
                self.Promise = None
            raise exc

        if self.Promise:
            self.Promise.complete((retcode, out, err))
            self.Promise = None
        return retcode, out, err

    def start(self):
        self.Promise = Promise(self.Tag)
        PyThread.start(self)
        return self.Promise
        
    @synchronized
    def killme(self):
        if self.Subprocess is not None and self.Subprocess.returncode is None:
            self.Subprocess.terminate()
            self.Killed = True
        
class ShellCommand(Subprocess):

    def __init__(self, command, tag=None, cwd=None, env=None, input=None, daemon=None, timeout=None):
        Subprocess.__init__(self, command, tag=tag,
            input=input, timeout=timeout, 
            shell=True, close_fds=True,
            stdout=subprocess.PIPE, 
            stderr=subprocess.PIPE, 
            stdin=None if input is None else suprocess.PIPE,
            cwd=cwd, env=env, daemon=daemon)

    @staticmethod
    def execute(command, cwd=None, env=None, input=None, timeout=None, daemon=True):
        s = ShellCommand(command, cwd=cwd, env=env, input=input, daemon=daemon, timeout=timeout)
        return s.run()
    
        